
/***********************************************************************
* SCCS id: @(#)iff_to_cdf.c	1.30 13:27:31 04/28/00 
************************************************************************
* PROJECT: Cluster, UK CDHF
*
* COMPONENT: Pipeline Processing, QIFF
*
* MODULE: iff_to_cdf, file conversion.
*
* LANGUAGE: ANSI C
*
* PURPOSE:  Creates a cdf file using a CSDS prototype cdf file & iff file.
*
* DESCRIPTION: Converts ascii data interface file from .iff syntax into cdf
*        file in CSDS standard. It uses a csds standard cdf file prototype 
*        generated from csds skeleton file as template. It uses the cdf file
*        to interpret the data in the iff file, which are then transferred
*        to the cdf file. The resulting cdf file is in CSDS standard format.
*
*
* FUNCTIONS:
*  QiIFFtoCDF
*  QiErrorHandler
*  QiStatusHandler
*  QiGetDay0
*  QiFNameParser
*  QiComposeName
*  QiLineParser
*  QiWriteHalfInterval
*  QiVersionStamp
*  QiSetupCDF
*  QiInitRecords
*  QiIFFScan
*  QiIFFRead
*  QiCtoISO
*  QiFillLine
*  QiPutGattr
*  QiGetVarInfo
*  QiWriteEpochAttrs
*  QiInterval
*
* MODULES_USED: 
*  cdf Library
*
* NOTES:
*
*  Uses header IFF_to_CDF.h for definitions, to attach libraries,
*  define function and structure prototypes.
*
*  cdf file, iff and log file names are derived from the generic file name 
*  in the argument list,
*  Cn_RR_III_DATE, as Cn_RR_III_DATE_V00.cdf, Cn_RR_III_DATE.iff and 
*  Cn_RR_III_DATE.log
*  
*  For use within the UK CDHF routine pipeline data reduction for
*  Cluster Prime and Summary Parameter data.
*
*
* COPYRIGHT:
* ---------------------------------------------------------------------
* Copyright (c) 1994 A.J.Allen
*
* Astronomy Unit, Queen Mary and Westfield College, London, UK.
* This software is is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
* If you are associated with the Cluster project, you may copy, or
*   modify this software freely, solely for purposes associated with
*   the Cluster project, and provided that you maintain a complete
*   copy of this copyright message with the software.
* If this software is distributed to you at the same time as a separate
*   file called COPYRIGHT, then the terms of copyright listed in that
*   file apply to this software.
* ---------------------------------------------------------------------
*
*
* HISTORY:
*
* V1.0   24 Nov 94    A.J.Allen  Pre-release version for CSDS-UI support
* V1.1   01 Dec 94    A.J.Allen  First Full Release
* V1.2   05 Dec 94       ''      Add CDF_INT2 & CDF_UINT2 (after P.Torrente)
* V1.3   4  Jan 95       ''      Reworked to allow use as a function call.
* V1.4   6  Feb 95       ''      Replace exit(1) with longjmp on ERROR
*                                Add option to set logfile path.
*                                Add traps for bad input lines.
* V1.5   23 Mar 95       ''      Fix Half_interval for unmarked data gaps.
* V1.6      Apr 95       ''      Remove isdigit on reading integers.
* V1.7                   ''      Changes of defaults for ampte use
* V1.8                   ''      Changes of defaults for ampte use
* V1.9   23 May 95       ''      Allow multiple element Vattr input.
*                                Allow hex escape sequences on input for chars
*                                Add Int2 Vattr data type
* V1.10  30 June 97      ''      Allow input of start time and data periodicity
*                                Public release version.
* V1.11  30 June 99      ''      Remove Data_intervals attribute post .skt V1.9 changes
*                                for Cluster II.
*
***********************************************************************/

#include "IFF_to_CDF.h"

/***********************************************************************
* SCCS id: @(#)iff_to_cdf.c	1.30 13:27:31 04/28/00 
************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  iff_to_cdf
*
* LANGUAGE: ANSI C
*
* FUNCTION: main
*
* PURPOSE: Main function of iff_to_cdf module.
*
* DESCRIPTION: Read input args, open files and call processing functions.
*
* ARGUMENTS:  
*
*  IN:         path to directory containing .iff file, e.g. /data/IFF/FGM
*  IN:         path to directory containing empty .cdf file, 
*              e.g. /V00data/FGM
*  IN:         generic file name in the form CX_RR_III_YYYYMMDD
*              which identifies spacecraft, resolution, instrument and date
*              e.g. C1_PP_FGM_19961225 is for Cluster 1, prime parameter
*              fluxgate magnetometer on 25 December 1996.
*  IN: Optional -nodelete flag. Indicates that cdf file is not to be deleted
*              if error is detected. Otherwise trapped errors result in the
*              output cdf file being deleted before exiting.
*  IN: Optional -log flag. Next argument is path to directory for log file
*
* FILES:
*
* IN:          .iff interface file with name derived from the argument list,
*              e.g.  CX_RR_III_YYYYMMDD.iff
*
* OUT:         .cdf file with name derived from the argument list,
*              e.g.  CX_RR_III_YYYYMMDD_V00.cdf
*
* OUT:         .log file with name derived from the argument list,
*              e.g.  CX_RR_III_YYYYMMDD.log
*
*
* GLOBALS:  FILE *logfile
*           long CDF_delete
*
* RETURN:
*    TYPE:   int
*    VALUES: 0   Successful completion
*            1   Fatal error detected
*
* FUNCTIONS_USED:
*         QiIFFtoCDF
*         QiErrorHandler
*         QiFNameParser
*
* ALGORITHM:
*       Test argument for option flag
*       Construct path and file names from arguments
*       Parse generic file name for later use of components
*       Call QiIFFtoCDF to do actual processing
*
* HISTORY:
*
*   See also module prologue.
*
*        1 Dec 94    A.J.Allen
*        First Release
*
***********************************************************************/

int main(int argc, char *argv[])

{

/* default log path is current WD, and log_fname is a global set in .h */ 
char log_path[FULL_NAME_LEN] = ".";  
char cdf_path[FULL_NAME_LEN], cdf_fname[FULL_NAME_LEN];
char iff_path[FULL_NAME_LEN], iff_fname[FULL_NAME_LEN];
char option[WORD_LEN];

struct QiSFName QiSname;
long check;


 /* process argument list */

 if(argc < 4) {
   fprintf (stderr,
      "syntax is\n fill_CDF [-flags] .iff_path .cdf_path .iff_file(s) \n");  
   exit(1);
 }
 
 /* test for option flags */
  ++argv;--argc;
  strcpy(option, *argv);
  while(option[0] == '-'){
    if(strcmp(option,"-nodelete") == 0)  CDF_delete = NO_DELETE;  
    if(strcmp(option,"-log") == 0)  {
      ++argv;--argc;  
      strcpy(log_path, *argv);
    }
    ++argv;--argc;
    strcpy(option, *argv);
  }
 
 /* initialise paths to .iff and .cdf files */
 
 strcpy(iff_path, *argv); /* initialise path to .iff */
 strcat(iff_path, "/");
 ++argv;--argc;
 strcpy(cdf_path, *argv); /* initialise path to .cdf */
 strcat(cdf_path, "/");
  
  ++argv;
  strcpy(QiSname.fname, *argv);
  strcpy(iff_fname, iff_path);strcat(iff_fname, QiSname.fname);
  strcat(iff_fname,".iff");
  strcpy(cdf_fname, cdf_path);strcat(cdf_fname, QiSname.fname);
  strcat(cdf_fname,"_V00");
  strcpy(log_fname, log_path); strcat(log_fname,"/");
  strcat(log_fname, QiSname.fname); strcat(log_fname,".log");
  
    /* parse file name root for later use of sub-strings and date */
    
  check = QiFNameParser(&QiSname);
    if(check != QMW_OK){
      fprintf(logfile,"Cannot parse input file name %s", QiSname.fname);
      exit(1);
    }
  check = QiIFFtoCDF(iff_fname, cdf_fname, log_fname, &QiSname);
    if(check != QMW_OK){
      fprintf(logfile,"Error returned from QiIFFtoCDF controller\n");
      exit(1);
    }

      exit(0);
              
}        /* end: main */

/***********************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  iff_to_cdf
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiIFFtoCDF
*
* PURPOSE: Open i/o files and launch processing functions.
*
* DESCRIPTION: This function performs the operations that would normally be 
*              done by the main sub-program. It opens files and launches the
*              processing functions. It is separated from mani to allow this 
*              module to be run as a stand alone process (with simple main) 
*              or as a linked callable function.
*
* ARGUMENTS:
*    IN:  char *iff_fname
*    IN:  char *cdf_fname
*    IN:  char *log_fname
*    IN:  struct QiSFName *QiSname
*
* GLOBALS:  FILE *logfile
*
* RETURN:
*    TYPE:   void
*    VALUES: none
*
* FUNCTIONS_USED:
*         QiErrorHandler
*         QiStatusHandler
*         QiFNameParser
*         QiSetupCDF
*         QiIFFScan
*         QiInitRecords
*         QiVersionStamp
*         QiIFFRead
*         QiWriteHalfInterval
*         QiWriteEpochAttrs
*         CDFlib
*
* ALGORITHM:
*       Open .log file for writing
*         If unable, set log file pointer to point to stderr
*       Open cdf file provided by control script
*       Open .iff file for reading
*       Parse generic file name for later use of components
*       Set up cdf quantities using cdf file
*       Scan through iff file doing counts and checks
*       Initialise enough space in prototype cdf to take records 
*       found in iff file
*       Write software and date stamps into cdf file
*       Read lines from iff file and write data to cdf file
*       Calculate and write Half_interval variable
*       Write epoch Vattributes for MIN and MAX values
*       Close cdf file
*
* HISTORY:
*
*   See also module prologue.
*
*        2 Jan 95    A.J.Allen
*        First Release, issue 2 added to facilitate use by ISDAT
*
***********************************************************************/

long QiIFFtoCDF (char *iff_fname,
                 char *cdf_fname,
                 char *log_fname,
                 struct QiSFName *QiSname)
 {

/* CDF   variables */

CDFid id;               /* ID of CDF for subsequent operations */
CDFstatus status;

long check=QMW_OK;
double data_res=60000; /* set to 1 min if not read */

struct QiSParams QiScdf;
struct QiSBuffer QiSrec;
struct QiSIO QiSio;
struct QiSInterval QiShi;
            
FILE *f;

/* initialise lines to NULL to allow test before free */
QiSio.line = NULL;
QiSio.line_cpy = NULL;
QiSrec.buff = NULL; 
QiShi.ivl_array = NULL;

  /* set longjump stop for Fatal Error return */
  /* Avoids use of exit on fatal error which cause parent module to exit */

  if( setjmp(jmp_env) != 0 ){
    QiSio.line = QiStrFree(QiSio.line);
    QiSio.line_cpy = QiStrFree(QiSio.line_cpy);
    if(QiSrec.buff != NULL) free(QiSrec.buff);
    if(QiShi.ivl_array != NULL) free (QiShi.ivl_array);
    return QMW_ERROR;  /* return to main with error state */
  }
  
  /*  open .log file */

  logfile = fopen(log_fname, "w" );
  if(logfile == NULL ){ /* attach log output to standard error */
     logfile = stderr;
     QiErrorHandler(QMW_WARNING, 
                    "Failed to attach log file for writing, using stderr");
  };

  /* open .cdf file */
 
   fprintf(logfile, "Opening cdf: %s.cdf\n", cdf_fname);
   status = CDFlib(OPEN_, CDF_, cdf_fname, &id, NULL_);
     QiStatusHandler(status, "problem opening cdf file" );
  
  /*  open .iff ASCII interface file */

  fprintf(logfile, "Opening iff: %s\n",iff_fname);
  f = fopen(iff_fname, "r" );
  if( f==NULL ){
    QiErrorHandler(QMW_ERROR, "Failed to open iff file.");
  };
       
   /* Find out about the .cdf file to be written into */
     
   check = QiSetupCDF(QiSname, &QiScdf, &QiSrec);
     QiErrorHandler(check,"Problem with prototype cdf file");
            
   /* READ .iff File */
       
   check = QiIFFScan(f, QiSname, &QiScdf, &QiSio, &data_res);
     QiErrorHandler(check,"Problem with first scan of .iff file");

   /* initialise records for all record varying variables. */
   
   check = QiInitRecords(&QiScdf, &QiShi);
     QiErrorHandler(check,"problem initialising space for variables");
                 
   /* set some global attributes that do not depend on .iff file */
   
   check = QiVersionStamp();
     QiErrorHandler(check,"Status returned from QiVersionStamp");
     
   /* Second pass to process the .iff file  */   
              
   check = QiIFFRead(f, &QiScdf, &QiSrec, &QiSio, &QiShi, data_res);
     QiErrorHandler(check,"Problem with second scan of .iff file");
     
   /* Having written the data we must now add the Half_interval variable */
   
   check = QiWriteHalfInterval(QiSname, &QiShi, QiScdf.n_recs, data_res);
     QiErrorHandler(check,"Half Interval not written");
     
   /* Write SCALE and VALID V-attributes for Epoch variable */
   
    check = QiWriteEpochAttrs(QiSname, &QiScdf);
     QiErrorHandler(check,"Status returned from QiWriteEpochAttrs");

   /* Ensure Inst_mode Global is filled */
   
    check = QiCleanUp(QiSname, &QiScdf);
     QiErrorHandler(check,"Status returned from QiCleanUp");
     
   /* Close CDFs and exit cleanly */
 
   fprintf(logfile,"\n Closing CDF %s\n\n",cdf_fname); 

   status = CDFlib(CLOSE_, CDF_, NULL_);
     QiStatusHandler(status,"Status returned from closing CDF");

   QiSio.line =  QiStrFree (QiSio.line);
   QiSio.line_cpy = QiStrFree(QiSio.line_cpy);
   if(QiSrec.buff != NULL) free(QiSrec.buff);
   if(QiShi.ivl_array != NULL) free (QiShi.ivl_array);
   
   return check;
}

/***********************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  iff_to_cdf
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiStatusHandler
*
* PURPOSE: Process response status of CDF library function calls.
*
* DESCRIPTION: Determine status returned by CDF library calls. Use returned 
*              status to determine nature of warnings/errors and initiate 
*              error processing when appropriate.
*
* ARGUMENTS:
*    IN:  CDFstatus  status
*    IN:  char       *note
*
* GLOBALS:  FILE *logfile
*
* RETURN:
*    TYPE:   void
*    VALUES: none
*
* FUNCTIONS_USED:
*         CDFerror
*         QiErrorHandler
*
* ALGORITHM:
*       If status is OK return. 
*       Otherwise output note and message returned from CDFerror to log file. 
*       Pass warning or error to QiErrorHandler.
*
* HISTORY:
*
*   See also module prologue.
*
*        1 Dec 94    A.J.Allen
*        First Release
*
***********************************************************************/

void QiStatusHandler (CDFstatus status, char *note)

{
  char message[CDF_STATUSTEXT_LEN+1];
  long check;

  if (status == CDF_OK){
    /* do nothing */
    return;
  }
  
  /* else print location note then fetch CDF message and print */
  
  fprintf(logfile,"%s\n", note);

  if (status < CDF_WARN){
     check = QMW_ERROR;
  }
  else
     if (status < CDF_OK){   
        fprintf (logfile,"Warning message from CDF...\n");
        check = QMW_WARNING;
     }
     else
     {
        fprintf (logfile, "CDF function executed OK, but...\n");
        check = QMW_OK;
     }      
   
     CDFerror (status, message);
     fprintf (logfile, " CDF> %s\n",message);
     QiErrorHandler(check, "\n");
     
     return;
}

/***********************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  iff_to_cdf
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiErrorHandler
*
* PURPOSE: Error handling.
*
* DESCRIPTION: Writes warning and error messages to logfile and
*              quits application on detected error. 
*
* ARGUMENTS:
*    IN:  long  status
*    IN:  char  *note
*
* GLOBALS:  FILE *logfile
*           long  CDF_delete
*
* RETURN:
*    TYPE:   void
*    VALUES: none
*
* FUNCTIONS_USED:
*         CDFlib
*
* ALGORITHM:
*     If status is OK return. 
*     Otherwise output note to log file. 
*     ERROR & WARNING lines have distinct prefixes to allow grep by 
*     control script.
*     If status is ERROR delete active cdf file (unless NO_DELETE flag set).
*     long jump to top level if status is ERROR. This will then return
*     ERROR state to calling module. Exit(1) is avoided as this would
*     cause calling module to exit too.
*
* HISTORY:
*
*   See also module prologue.
*
*        1 Dec 94    A.J.Allen
*          First Release
*        6 Feb 95    A.J.Allen
*          replaced exit(1) with longjmp
*
***********************************************************************/

void QiErrorHandler (long status, char *note)

{
  if (status == QMW_OK)  return;
    
  fprintf(logfile,"\n=> %s\n", note);

  if (status == QMW_ERROR){
  
    fprintf (logfile, "==> Fatal error: logged in %s\n", log_fname);
    if(CDF_delete != NO_DELETE){
      fprintf (logfile, "==> Deleting CDF file...\n");
      CDFlib(DELETE_, CDF_, NULL_);
    }
    longjmp(jmp_env, -1); /* return to top level to bale out cleanly */
  }
  else{
        fprintf (logfile, "=> Warning from iff_to_cdf \n");
  }
   
     return;
}

/***********************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  iff_to_cdf
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiLineParser
*
* PURPOSE: Parse input line into words.
*
* DESCRIPTION: Breaks input line into tokens (words). 
*              Substrings delimited by the character DELIM are treated as
*              a single word. Otherwise each word is separated by a space 
*              character or newline character. The separators and DELIM 
*              are stripped from the words found. The words are copied into 
*              an array of strings to protect from accidental corrupton of 
*              input line. This array is reused on subsequent calls.
*
* ARGUMENTS:
*    IN/OUT:  struct QiSIO *QiSio   Temporary space for input line and its 
*                                   substrings.
*
* GLOBALS:  
*
* RETURN:
*    TYPE:   long
*    VALUES: number of words found (quoted strings are treated as 
*             single words).
*            0 if the line is NULL or contains no non-space or newline 
*             characters.
*
* FUNCTIONS_USED:
*         CDFlib
*         QiErrorHandler
*
* ALGORITHM:
*        replace spaces with value BLOCK inside strings quoted with DELIM.
*        Use strtok to parse into tokens on remaining  space characters. 
*        Copy each token to successive entries in the array of strings 
*        called words.
*        Remove quotes and turn BLOCK to space character in all words
*        with DELIM as first and last character.
*        NULL quoted strings are set to a single space character to avoid 
*        write error from CDFlib. e.g., if DELIM=" then `""' becomes ` '.
*       
*
* HISTORY:
*
*   See also module prologue.
*
*        1 Dec 94    A.J.Allen
*        First Release
*
***********************************************************************/

long QiLineParser(struct QiSIO *QiSio) 
{
  char *separators ={" \n"}; /* space or new line */
  int n, n_char;
  char  *token;
  

/* note that `words' is an array of strings */

/* make a copy for parsing so that we can keep original to print warnings */

  strcpy(QiSio->line_cpy, QiSio->line);             
                
  /* Mask spaces in quoted strings */
    n_char = 0;
    while(QiSio->line_cpy[n_char] != '\0'){
      if(QiSio->line_cpy[n_char] == DELIM){
        n_char++;
        /* shift space character if present to avoid parsing */
        for(;QiSio->line_cpy[n_char] != DELIM; n_char++){
          if(QiSio->line_cpy[n_char] == ' ') QiSio->line_cpy[n_char] = BLOCK;
          if(QiSio->line_cpy[n_char] == '\n') {
            QiErrorHandler(QMW_WARNING, "incomplete line");
            return 0;    
          }
        }
      }
      n_char++;
    }

  /* start parsing and trap null input line */
   if ((token = strtok(&(QiSio->line_cpy[0]), separators)) == NULL) return 0;
   
   strcpy(&(QiSio->words[0][0]), token);
  /* split line into words */
  
  for(n=1; (token=strtok(NULL,separators)) != NULL; n++){
    if(n == NUMBER_OF_WORDS){
      QiErrorHandler(QMW_ERROR,"exceeded NUMBER_OF_WORDS in line parser");
    }
    if ((int) strlen(token) > WORD_LEN - 1) {
      fprintf(logfile, "\n\n%s\n\n",token);
      QiErrorHandler(QMW_WARNING,
                 "Word or quoted string too long, truncated...");
    }
    strncpy(&(QiSio->words[n][0]),token, WORD_LEN);
    
   /* add NULL as strncpy will not do so when token is too long */
    QiSio->words[n][WORD_LEN - 1] = '\0';

  /* remove character string delimiters DELIM, first word cannot be string */
       
    if(token[0] == DELIM && token[strlen(token)-1] == DELIM){ 
    
      if (strlen(token) == 2){ 
        /* set null strings to white space for cdf Gattr write */
        strcpy(&(QiSio->words[n][0]), " ");
      }
      else{
        /* rewrite words[n] without first quote */
        strcpy(&(QiSio->words[n][0]), &token[1]);
      
        /* overwrite end string delimiter */
        QiSio->words[n][strlen(QiSio->words[n])-1] = '\0'; 
      
        /* Put back space chars in quoted strings */
        for(n_char=0; n_char<(int)strlen( &(QiSio->words[n][0]) ); n_char++){
          if(QiSio->words[n][n_char] == BLOCK) QiSio->words[n][n_char] = ' ';
        } /* end for on n_char */  
      }
    } /* end if on DELIM */
   

  } /* end for over words */
  
  /* return number of words , ie 0 to m = m+1 = n */
  QiSio->n_words = n;
  return n;
 
} /* end QiLineParser */

/**********************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  iff_to_cdf
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiFNameParser
*
* PURPOSE: Parse input generic file name into identifiers.
*
* DESCRIPTION: Breaks input file name root into:-
*              spacecraft (e.g. C3 or CL etc)
*              resolution, SP or PP
*              instrument 3char abbreviaton
*              date string and year/month/day values
*
* ARGUMENTS:
*    IN/OUT:  struct QiSFName *QiSname    Structure for filename & 
*                                         substrings/values.
*
* GLOBALS:  
*
* RETURN:
*    TYPE:   long
*    VALUES: QMW_OK      Executed successfully.
*            QMW_ERROR   Fatal error detected.
*
* FUNCTIONS_USED:
*         
*
* ALGORITHM:
*        Make a copy of fname to allow safe use of strtok.
*        Parse copy on "_" and set pointers to appropriate substrings.
*        Mask off extension beyond required 8 chars of date substring 
*        using '\0'.
*        Evaluate year/month/day according to position in date substring
*        from back using '\0' to cut off each piece.
*
* HISTORY:
*
*   See also module prologue.
*
*        1 Dec 94    A.J.Allen
*        First Release
*
***********************************************************************/

long QiFNameParser(struct QiSFName *QiSname)
{
  
  char *separators ={"_"};
  char *ptr;
  long check=QMW_OK;
 
 /* this function is called only by the pipeline processing software, 
    so it sets the optional filename extension to null. This null is used
    in variable name construction to choose to use either __sc_rr_iii or 
    __extn extensions.      
 */
 
  strcpy(QiSname->sub_strings, QiSname->fname);  /* avoid messing up fname */
   
  QiSname->sc  = strtok(QiSname->sub_strings, separators);
  if(QiSname->sc == NULL) check = QMW_ERROR;
  
  QiSname->rr  = strtok(NULL,separators);
  if(QiSname->rr == NULL) check = QMW_ERROR;
  
  QiSname->iii = strtok(NULL,separators);
  if(QiSname->iii == NULL) check = QMW_ERROR;
  
  ptr = strtok(NULL,separators);
  if(ptr == NULL) check = QMW_ERROR;
  
  if((int)strlen(ptr) < 8) check = QMW_ERROR;
  else{
    /* strip off day, month, year from back */
    ptr[8] = '\0'; /* just in case name is extended */
    QiSname->day = atoi(ptr+6);
    *(ptr+6) = '\0';
    QiSname->month = atoi(ptr+4);
    *(ptr+4) = '\0';
    QiSname->year = atoi(ptr);
  }
  
  QiSname->extn = NULL;  /* set extension to null for pipeline processing */

   return check;
 
} /* end QiFNameParser */

/**********************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  iff_to_cdf
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiGetDay0
*
* PURPOSE: Find epoch msec value of start of day.
*
* DESCRIPTION: Determine the epoch (as double msec) for the start of 
*              the day specified.
*
* ARGUMENTS:
*    IN:  long year    year element of date
*    IN:  long month   month element of date
*    IN:  long day     day element of date
*
* GLOBALS:  
*
* RETURN:
*    TYPE:   double
*    VALUES: msec value in cdf epoch for midnight at start of specified day.
*
* FUNCTIONS_USED:
*         QiErrorHandler
*         computeEPOCH
*
* ALGORITHM: pass date and time=0 to cdf routine and inspect status of reply.
*        
* HISTORY:
*
*   See also module prologue.
*
*        1 Dec 94    A.J.Allen
*        First Release
*
***********************************************************************/

double QiGetDay0(long year, 
                 long month, 
                 long day)
{
  double msec_day_0;
  
  msec_day_0 = computeEPOCH(year, month, day, 0, 0, 0, 0);
  if(msec_day_0 < 0){
    QiErrorHandler(QMW_ERROR, 
                   "error from computeEPOCH for time at start of day");
  }
  
  return msec_day_0;
 
} /* end QiGetDay0 */

/**********************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  iff_to_cdf
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiVersionStamp
*
* PURPOSE: Write metadata that does not come from .iff file.
*
* DESCRIPTION: Some metadata is not available in standard csds files, 
*              but must be determined at the time of populating the cdf file. 
*              Some of this metadata is determined and written here.
*
* ARGUMENTS:
*   
*
* GLOBALS:  
*
* RETURN:
*    TYPE:   long
*    VALUES: QMW_OK      Executed successfully.
*            QMW_ERROR   Fatal error detected.
*
* FUNCTIONS_USED:
*         CDFlib
*         QiErrorHandler
*         QiStatusHandler
*         QiCtoISO
*
* ALGORITHM: 
*     Write the Generated_by Gattr specified in the header file.
*     Write the Software_version identifier of this software.
*     Extract and write the Software version of the CDF library being used.
*     Determine the date/time from platform clock and convert to Iso Standard
*     Write time string as Generation_date Gattr.
*     Write SCALEMIN, SCALEMAX, VALIDMIN, VALIDMAX for Epoch variable
*     as start and end of day being handled.
*        
* HISTORY:
*
*   See also module prologue.
*
*        1 Dec 94    A.J.Allen
*        First Release
*
***********************************************************************/

long QiVersionStamp()
{
  
  CDFstatus status;
  char cdf_ver[WORD_LEN];
  time_t tnow;
  char* today;
  long ver, rel, inc;  
  long check=QMW_OK;
  char iso_time[25];

  
  /* Generated_by */
  
  status = CDFlib(SELECT_, ATTR_NAME_,  "Generated_by", NULL_);
     QiStatusHandler(status,"can't locate G attribute `Generated_by' " );

  CDFlib(SELECT_, gENTRY_, 0L,   NULL_);        
  status = CDFlib(PUT_, gENTRY_DATA_, CDF_CHAR, (long)strlen(GEN_BY), GEN_BY, 
                  NULL_);
     QiStatusHandler(status,"failed to add entry to `Generated_by' " );
  
  /* Software_version */
  
  status = CDFlib(SELECT_, ATTR_NAME_,  "Software_version", NULL_);
     QiStatusHandler(status,"can't locate G attribute `Software_version' " );
  
  /*  QMW sw version from header IFF_to_CDF.h */
  CDFlib(SELECT_, gENTRY_, 0L,   NULL_);        
  status = CDFlib(PUT_, gENTRY_DATA_, CDF_CHAR, (long)strlen(SW_VER), SW_VER,
                  NULL_);
     QiStatusHandler(status,
                     "failed to add QMW sw version to `Software_version' " );

  /* cdf version  */
  CDFlib(GET_, CDF_VERSION_, &ver, NULL_);
  CDFlib(GET_, CDF_RELEASE_, &rel, NULL_);
  CDFlib(GET_, CDF_INCREMENT_, &inc, NULL_);
  sprintf(cdf_ver,"CDF version %ld.%ld.%ld", ver, rel, inc);
  
  CDFlib(SELECT_, gENTRY_, 1L,   NULL_);        
  status = CDFlib(PUT_, gENTRY_DATA_, 
                              CDF_CHAR, (long)strlen(cdf_ver), &cdf_ver[0],
                  NULL_);
     QiStatusHandler(status,
                     "failed to add cdf version to `Software_version' \n" );

  /* Generation_date */
  
   time(&tnow);
   today = QiCtoISO(ctime(&tnow), iso_time);
   
  status = CDFlib(SELECT_, ATTR_NAME_,  "Generation_date", NULL_);
     QiStatusHandler(status,"can't locate G attribute `Generation_date' " );

  CDFlib(SELECT_, gENTRY_, 0L,   NULL_);        
  status = CDFlib(PUT_, gENTRY_DATA_, CDF_CHAR, (long)strlen(today), today, 
                  NULL_);
     QiStatusHandler(status,"failed to add entry to `Generation_date' " );
  
  return check;
 
} /* end QiVersionStamp */

/***********************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  iff_to_cdf
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiWriteHalfInterval
*
* PURPOSE: Write Half interval variable to cdf file.
*
* DESCRIPTION: The half interval is determined during the preliminary scan
*              of thi iff file. It is written to the cdf file here.
*
* ARGUMENTS:
*    IN:   struct QiSFName *QiSname    Structure for filename & 
*                                      substrings/values.
*    IN:  struct QiSInterval *QiShi    structure for data interval quantities
*    IN:  long   num_records           Total number of records in the file.
*
* GLOBALS:  
*
* RETURN:
*    TYPE:   long
*    VALUES: QMW_OK     Executed successfully.
*            QMW_ERROR  Fatal error detected.
*
* FUNCTIONS_USED:
*         CDFlib
*         QiErrorHandler
*         QiStatusHandler
*         QiComposeName
*
* ALGORITHM: 
*    Construct full Half_interval variable name for file and select.
*    If SP file write fixed half minute value.
*    If PP file
*       If intervals are uniform to tolerance
*         write average half_interval
*       Else
*          set to record varying
*          halve values in interval array
*          write this array of half_interval values
*    Write Half_interval CATDESC attribute entry
*        
* HISTORY:
*
*   See also module prologue.
*
*        1 Dec 94    A.J.Allen
*        First Release
*
***********************************************************************/

long QiWriteHalfInterval(struct QiSFName *QiSname, 
                         struct QiSInterval *QiShi,
                         long num_records,
                         double data_res)
{
 double Half_ivl;
 CDFstatus status;
 char varname[CDF_VAR_NAME_LEN+1];
 long check=QMW_OK;
 char attr_text[WORD_LEN];
 long attr_n_elem;
 char txt[TEXT_LEN];
 long n;
 
 /* construct variable name */
 
 check = QiComposeName(varname, "Half_interval", QiSname);
 QiErrorHandler(check, "failed to construct variable name");
 
 /* select variable */
 
 status = CDFlib(SELECT_, zVAR_NAME_, varname, NULL_);
   sprintf(txt,"failed to find variable = %s ",varname);
   QiStatusHandler(status, txt);
 
 /* set from data_res */
 
       Half_ivl = data_res * 0.5;

  if(QiSname->rr[0] == 'S' || QiSname->rr[0] == 's'){ /* SPDB */
   
    /* This is always non-record varying as it is a fixed interval */
   
 
    /* enter as first record */
    status = CDFlib(SELECT_, zVAR_RECNUMBER_, 0L, NULL_);
   
    status = CDFlib(PUT_, zVAR_DATA_, (void *)&Half_ivl, NULL_);
     QiStatusHandler(status,
                     "failed to add entry to `Half_interval' variable " );
   
  }
  else { /* PPDB (or other) */ 
     
   /* If no interval can be measured use value read */
   if( QiShi->ivl_num < 1 ) {

        /* enter as first record */
      status = CDFlib(SELECT_, zVAR_RECNUMBER_, 0L, NULL_);
   
      status = CDFlib(PUT_, zVAR_DATA_, (void *)&Half_ivl, NULL_);
         QiStatusHandler(status,
                     "failed to add fillval to `Half_interval' variable " );
   
      return QMW_OK;
    }
   
    /* select first record */
    status = CDFlib(SELECT_, zVAR_RECNUMBER_, 0L, NULL_);

    /* If non-record varying write once */
    
    if(QiShi->ivl_max - QiShi->ivl_min < IVL_TOLERANCE){   
     status = CDFlib(PUT_, zVAR_DATA_, (void *)&(QiShi->ivl_ave), NULL_);
       QiStatusHandler(status,
                       "failed to add entry to `Half_interval' variable ");
    }
    else{
   
     /* set to record varying */  
     status = CDFlib(PUT_, zVAR_RECVARY_, VARY, NULL_);
       QiStatusHandler(status,
                      "failed to change `Half_interval' variable to VARY\n");
          
     /* set record count to number of records to write */
     status = CDFlib(SELECT_, zVAR_RECCOUNT_, num_records, NULL_);
          
     /* set record interval to every record */
     status = CDFlib(SELECT_, zVAR_RECINTERVAL_, 1L, NULL_);
   

     /* write array of Half_interval values */
     
     /* first halve the interval values */
     for(n=0; n<num_records ; n++){
       QiShi->ivl_array[n] = QiShi->ivl_array[n] * 0.5;
     }
     status = CDFlib(PUT_, zVAR_HYPERDATA_, 
                                      (void *)&(QiShi->ivl_array[0]), NULL_);
       QiStatusHandler(status,"failed to write `Half_interval' data array ");
     
    } /* end if on Variance */
  } /* end if on SPD/PPD */
 
 /* write out variable attributes needed */

 status = CDFlib(SELECT_, zENTRY_NAME_, varname, NULL_);
   QiStatusHandler(status,"failed to select attribute entry for ");
 
/* CATDESC */
 
 strcpy(attr_text,
       "Half the mean separation between time tags, averaged over the file");
 
 status = CDFlib(SELECT_, ATTR_NAME_, "CATDESC", NULL_);
   QiStatusHandler(status,"failed to select attribute 'CATDESC' " );
 
 attr_n_elem = strlen(attr_text)+1;
 status = CDFlib(PUT_, zENTRY_DATA_, CDF_CHAR, 
                                     attr_n_elem, (void *)attr_text, NULL_);
   QiStatusHandler(status,"failed to write attribute 'CATDESC' " );
     
  return check; /* if no FATAL errors trapped by QiStatusHandler return OK */
 
} /* end QiWriteHalfInterval */

/***********************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  iff_to_cdf
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiSetupCDF
*
* PURPOSE: Read prototype CDF.
*
* DESCRIPTION: Reads the prototype cdf to determine the number and nature
*              of variables required to be populated. This determines the
*              expected structure of a record line in the input iff file.
*              Memmory is allocated for a record buffer for reading. The 
*              Parameters structure is assembled to allow future 
*              determination of record varying variable details by variable
*              number (e.g. dimension, sizes, data type).
*
* ARGUMENTS:
*    IN/OUT:   struct QiSFName *QiSname   Structure for filename & 
*                                         substrings/values.
*    IN/OUT:  struct QiSParams *QiScdf
*    IN/OUT:  struct QiSBuffer *QiSrec
*
* GLOBALS:  
*
* RETURN:
*    TYPE:   long
*    VALUES: QMW_OK     Executed successfully.
*            QMW_ERROR  Fatal error detected.
*
* FUNCTIONS_USED:
*         CDFlib
*         QiErrorHandler
*
* ALGORITHM:
*       Determine number of variables in cdf
*       For each variable:
*          Note variable number if the Epoch variable
*          Construct list of VARY variables
*          if VARY:
*             Find Dimension
*             Find sizes 
*             Increment number of entries expected in record
*             Construct total number of entries
*             use QiGetVarInfo to determine FILLVAL, data type and byte 
*             length
*             Find number of elements
*             Increment buffer size += Dim*size*number_elements
*       Allocate memory sufficient for full record.
*       Determine Epoch msec value at start of day based on date component
*       of file name.
*
* HISTORY:
*
*   See also module prologue.
*
*        1 Dec 94    A.J.Allen
*        First Release
*     
***********************************************************************/

long QiSetupCDF(struct QiSFName *QiSname, 
               struct QiSParams *QiScdf,
               struct QiSBuffer *QiSrec ){
   
   long varnum;
   long vary;
   long dim;
   long check = QMW_OK;
   char varname[CDF_VAR_NAME_LEN+1];
   CDFstatus status;
   char txt[TEXT_LEN];

   QiSrec->buff_len = 0;
   QiScdf->n_list = 0;
   QiScdf->n_data = 0;
   
   status = CDFlib(GET_, CDF_NUMzVARS_, &(QiScdf->n_vars), NULL_);
     QiStatusHandler(status," Problem getting number of variables" );
     
     /*  initialise information on each rec-varying variable */
     
     for( varnum = 0; varnum < QiScdf->n_vars; varnum++){
     
       status = CDFlib(SELECT_, zVAR_, varnum, NULL_);
         sprintf(txt, " Problem selecting variable number %ld", varnum);
         QiStatusHandler(status, txt);

       /* identify EPOCH  for later */
       status = CDFlib(GET_, zVAR_NAME_, varname, NULL_);
         sprintf(txt, "Problem getting name of variable number %ld", varnum);
         QiStatusHandler(status, txt);    

       if(strncmp(varname, "Epoch", 5) == 0){
          QiScdf->EPOCH_var_num = varnum;
       }
       
       status = CDFlib(GET_, zVAR_RECVARY_, &vary, NULL_);
         sprintf(txt, " Problem getting variance of variable %s", varname);
         QiStatusHandler(status, txt);
         
       QiScdf->recvary[varnum] = vary;
       if(vary == VARY ){
          
          /* establish list of VARY variables for assembling buffer later */
          QiScdf->list_vars[QiScdf->n_list] = varnum;
          QiScdf->n_list++;
                   
         /* determine dimension of each dimensioned variable */
         status = CDFlib(GET_, zVAR_NUMDIMS_, &dim, NULL_);
         sprintf(txt, " Problem getting dimension of variable %s", varname);
           QiStatusHandler(status, txt);
           
         QiScdf->var_dim[varnum] = dim;
         
         /* determine size of each dimensioned variable */
         if(dim > 0){
           status = CDFlib(GET_, zVAR_DIMSIZES_, 
                                      &(QiScdf->dim_size[varnum][0]), NULL_);
           sprintf(txt, " Problem getting size of variable %s", varname);
             QiStatusHandler(status, txt);
             
         } /* end if on dimension */
         else{
           QiScdf->dim_size[varnum][0] = SIZE_0;
         }
         
         /* find number of entries in variable */
         QiScdf->n_cells[varnum] = 1;
         for (dim = 0; dim < QiScdf->var_dim[varnum]; dim++){
           QiScdf->n_cells[varnum] = QiScdf->n_cells[varnum] * 
                                     QiScdf->dim_size[varnum][dim];
         }
         
         /* increment number of data entries expected per input line */
         
         QiScdf->n_data += QiScdf->n_cells[varnum];
         
        /* Find data type and FILLVAL */
        
          status = QiGetVarInfo(QiScdf, QiSrec, &(varname[0]), varnum);         
          sprintf(txt, 
                  "Can't get FILLVAL or Data Type of variable %s", varname);
          QiStatusHandler(status, txt);
                    
         /* find number of elements for each variable */
         
         status = CDFlib(GET_, zVAR_NUMELEMS_, &(QiScdf->n_chars[varnum]), 
                         NULL_);
           sprintf(txt, " Can't get num elements for variable %s", varname);
           QiStatusHandler(status, txt);
         
         /* increment buffer size to accomodate each variable */
         
          QiSrec->buff_len += QiScdf->n_chars[varnum] * 
                              QiScdf->n_cells[varnum] *
                              QiScdf->data_len[varnum]; 
                  
       } /* end if on variance */
     } /* end for over variables */
     
     /* allocate memory for buffer sufficient for full record */
     QiSrec->buff = (unsigned char *)malloc(QiSrec->buff_len); 
     
     /* set up Epoch for start of day */
     
     QiScdf->msec_day = QiGetDay0(QiSname->year, QiSname->month, QiSname->day);
     
     /* print as a test */
     /* encodeEPOCH(QiScdf->msec_day, estr); */
     /*  printf("%s\n",estr); */
          
      return check;
            
 }   /* end     QiSetupCDF */
 
/***********************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  iff_to_cdf
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiComposeName
*
* PURPOSE: Combines variable name root with standard file extension.
*
* DESCRIPTION: Constructs full variable name from an input root value and
*              the filename.
*
* ARGUMENTS:
*    OUT:  char *varname               Returned full  variable name  
*    IN:  char *stem                   Input specific name portion
*    IN:   struct QiSFName *QiSname    Structure for filename & 
*                                      substrings/values.
*
* GLOBALS:  
*
* RETURN:
*    TYPE:   long
*    VALUES: QMW_OK     Executed successfully.
*
* FUNCTIONS_USED:
*         
*
* ALGORITHM:
*            Copy stem to variable name
*            Catenate elements from file name structure and separators
*
* HISTORY:
*
*   See also module prologue.
*
*        1 Dec 94    A.J.Allen
*        First Release
*     
***********************************************************************/


long QiComposeName(char *varname, 
                 char *stem,
                 struct QiSFName *QiSname)
{
 
 /* construct variable name from stem and postfix */
  
 /* if optional extension extn is set this must be used */
 
 if(QiSname->extn != NULL){
     strcpy(varname, stem);
     strcat(varname, QiSname->extn);
   
   /* else add no extension */
 }
 /* otherwise use standard extension from filename */
 else{
   strcpy(varname, stem);strcat(varname, "__");
   strcat(varname, QiSname->sc);strcat(varname,"_");
   strcat(varname, QiSname->rr);strcat(varname,"_");
   strcat(varname, QiSname->iii);
 }
 
 return QMW_OK;
 
} /* end QiComposeName  */

/***********************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  iff_to_cdf
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiInitRecords
*
* PURPOSE: Initialise each variable in cdf to have contiguous memory 
*          sufficient for number of records in iff file.
*
* DESCRIPTION: Uses CDFlib to initialise each variable in turn to the
*              number of records found in iff file. CDFlib will try
*              to allocate contiguous memory for each variable.
*
* ARGUMENTS:
*    IN/OUT:  struct QiSParams *QiScdf    Structure containing cdf variable
*                                         quantities. 
*    IN/OUT:  struct QiSInterval *QIShi   Structure for half interval info
*
* GLOBALS:  
*
* RETURN:
*    TYPE:   long
*    VALUES: QMW_OK     Executed successfully.
*
* FUNCTIONS_USED:
*         QiStatusHandler
*         CDFlib
*
* ALGORITHM:
*            If iff file contains no records, set n_recs = 1.
*            malloc space for half interval variable
*            For each VARY variable
*               select variable
*               initialise to n_recs records (FILLVAL set to that of
*               prototype cdf)
*
* HISTORY:
*
*   See also module prologue.
*
*        1 Dec 94    A.J.Allen
*        First Release
*     
***********************************************************************/

long QiInitRecords(struct QiSParams *QiScdf,
                   struct QiSInterval *QiShi){

   long list;
   long check = QMW_OK;
   CDFstatus status;
   char txt[TEXT_LEN];
      
    /* Ensure that empty file has one record of fills */

    if (QiScdf->n_recs == 0) QiScdf->n_recs = 1;

    /* allocate space for Half-interval array */

    QiShi->ivl_array = (double *) malloc(QiScdf->n_recs*sizeof(double));

     /* initialise records for each record varying variable, 
         using list_vars  */
     
     for( list = 0; list < QiScdf->n_list; list++){
     
    /* select variable from list of record varying vars */
    
       status = CDFlib(SELECT_, zVAR_, QiScdf->list_vars[list], NULL_);
       if (status != CDF_OK) {
         sprintf(txt,
                  "Problem selecting variable %ld", QiScdf->list_vars[list]); 
         QiStatusHandler(status, txt);
         check = QMW_ERROR;
       }
                       
  /* initialise space allocation within file for each rec-varying variable */
    
       status = CDFlib(PUT_, zVAR_INITIALRECS_, QiScdf->n_recs, NULL_);
       if (status != CDF_OK) {
         sprintf(txt,
                 "Problem initialising var # %ld", QiScdf->list_vars[list]); 
         QiStatusHandler(status, txt);
         check = QMW_ERROR;
       }
                  
     } /* end for over variables */
         
      return check;
            
 }   /* end   QiInitRecords */
 
/***********************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  iff_to_cdf
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiIFFScan
*
* PURPOSE: First scan of iff file to establish number of records.
*
* DESCRIPTION: Reads each line of input iff file. Tests for Fatal error line.
*              Counts number of good and bad records.
*              Sets up fill value array if `U' line present.
*              Tests that Logical_file_id in prototype cdf is compatible 
*              with input arg.
*
* ARGUMENTS:
*    IN:       FILE *f,                        pointer to iff file
*    IN:        struct QiSFName *QiSname       Structure for filename & 
*                                              substrings/values.
*    IN/OUT:   struct QiSParams *QiScdf        structure with cdf quantities
*    IN:       struct QiSIO *QiSio             structure to hold input line
*
* GLOBALS:  
*
* RETURN:
*    TYPE:   long
*    VALUES: QMW_OK     Executed successfully.
*
* FUNCTIONS_USED:
*         QiErrorHandler
*         QiFillLine
*         QiLineParser
*
* ALGORITHM:
*            Read each iff file line in turn.
*            While line not NULL
*               If line is `fatal error' call QiErrorHandler to exit code 
*               cleanly
*               If line is a data record:
*                  If no EOD marker found increment record counter
*                  else increment excess record counter
*                  parse line to get number of words in line
*                  If number of words incompatible with expected line length
*                     Increment bad record counter
*               If line is Vattr
*                  Test filename from arg agrees with Logical_file_id in cdf
*               If line if Fill value line
*                   parse line
*                   write fill values from line into array of strings
*               If line is EOD:
*                   set EOD flag 
*                   If number of lines read not that of EOD line
*                     issue Warning 
*                     proceed using number of good lines found
*             Rewind file
*             issue warnings if bad lines or lines after EOD marker found               
*            
*
* HISTORY:
*
*   See also module prologue.
*
*        1 Dec 94    A.J.Allen
*        First Release
*     
***********************************************************************/
 
long QiIFFScan(FILE *f, 
         struct QiSFName *QiSname, 
         struct QiSParams *QiScdf,
         struct QiSIO *QiSio,
         double *data_res){

  long n_expct;
  long has_eod = NOT_FOUND;
  long check = QMW_OK;
  long n_over = 0;
  char txt[TEXT_LEN];
     
   /* initialise number of fill values */
   
   QiScdf->n_fills = 0;

   /* find number of records in .iff file to initialise. */
   
   QiScdf->n_recs = 0;
   
   while( QiFillLine(QiSio, f) != NULL ){

     if (dbg == 1) printf(">>%s\n", QiSio->line);
 
     switch(QiSio->line[0]){
     
     case 'F':  case 'f':
           
     {             /* Fatal error state */
     
        fprintf(logfile,"==> Fatal Error flag in .iff file...\n");
        fprintf(logfile,"==> %s",&(QiSio->line[0]) );
        QiErrorHandler(QMW_ERROR,
        "iff_to_cdf: Abort on fatal error flag passed from external module");
        break;
        
     } /* end case 'F' */
      
     case 'S':  case 's':  /* expected data periodicity */
           
     {             
     
       /* parse line into tokens */

       if(QiLineParser(QiSio) < 2)  { 
          fprintf(logfile, "\n%s\n", &(QiSio->line[0]));
          strcpy(txt, "syntax error in above Period line ");
          QiErrorHandler(QMW_WARNING, txt);
          check = QMW_WARNING;
          break;  /* skip this input line */
       }
       /* get data periodicity */
       
       *data_res = atof(&(QiSio->words[1][0]) );
       fprintf(logfile, "data res read as %g\n", *data_res);
       
       break;
        
     } /* end case 'S' */

     case 'I':  case 'i':
           
     {       /* variable name extension override */
     
       /* parse line into tokens */

       if(QiLineParser(QiSio) < 2)  { 
         fprintf(logfile, "\n%s\n", &(QiSio->line[0]));
         strcpy(txt, "syntax error in above var name extension override line ");
          QiErrorHandler(QMW_WARNING, txt);
          check = QMW_WARNING;
          break;  /* skip this input line */
       }
       /* set extension */
       
       QiSname->extn = malloc( strlen(&(QiSio->words[1][0])) + 1);
       strcpy(QiSname->extn,&(QiSio->words[1][0]) );
   
       if(strcmp(QiSname->extn, " ") == 0) strcpy(QiSname->extn, "");

       fprintf(logfile, "extn set to \"%s\"\n", QiSname->extn);
       
       break;
        
     } /* end case 'I' */

     case 'T':  case 't':
           
     {             /* override offset time */
     
       /* parse line into tokens */

       if(QiLineParser(QiSio) < 4)  { 
          fprintf(logfile, "\n%s\n", &(QiSio->line[0]));
          strcpy(txt, "syntax error in above time offset line ");
          QiErrorHandler(QMW_WARNING, txt);
          check = QMW_WARNING;
          break;  /* skip this input line */
       }
       /* override time offset */
       
     
       QiScdf->msec_day = QiGetDay0(atol(&(QiSio->words[1][0])), 
                                    atol(&(QiSio->words[2][0])), 
                                    atol(&(QiSio->words[3][0])) );
                                    
       fprintf(logfile, "Time offset %f\n", QiScdf->msec_day);

       break;
        
     } /* end case 'T' */
     
     case 'R': case 'r':
     
     {          
       /* count records  */
       if(has_eod == NOT_FOUND){
         (QiScdf->n_recs)++;
       }
       else{
         /* ignore data records after 'End of Data' marker */
         check = QMW_WARNING;
         n_over++;
       }
       
       /* count number of elements in record */
      if(QiLineParser(QiSio) == 0) {            
        sprintf(txt,"NULL line detected by QiLineParser\n");
        QiErrorHandler(QMW_ERROR, txt);
      }
       break;
     }
     
     case 'G': case 'g':
     
     {           /* Do consistency checks on Logical file id */
     
       /* parse line into tokens */

       if(QiLineParser(QiSio) < 3)  { 
          fprintf(logfile, "\n%s\n", &(QiSio->line[0]));
          strcpy(txt, "syntax error in above Gattribute line ");
          QiErrorHandler(QMW_WARNING, txt);
          check = QMW_WARNING;
          break;  /* skip this input line */
       }
       /* check file name */
       if( strcmp(&(QiSio->words[1][0]), "Logical_file_id") == 0){
         if(!strncmp(&(QiSio->words[2][0]), 
                     QiSname->fname, strlen(QiSname->fname) ) == 0){
           sprintf(txt,".iff file name not consistent with Logical_file_id");
           QiErrorHandler(QMW_WARNING, txt);
         }
       } /* end if on filename test */
       break;
      } /* end case 'G' */
      
      case 'U': case 'u':
      {  /* set fill values expected from iff file if present */
         /* Note: fill values to be used in cdf file are taken from cdf */
         
       int n;
        
        /* parse line into tokens */ 
        QiScdf->n_fills = QiLineParser(QiSio) - 3; 
        if(QiScdf->n_fills != QiScdf->n_data){
           strcpy(txt,"Wrong number of fillvalues found in 'U' line");
           QiErrorHandler(QMW_WARNING, txt);
           check=QMW_WARNING;
           break;
        }
        
        /* write new iff input fill values into fill array */
        for(n=0; n < QiScdf->n_fills; n++){
          strcpy(QiScdf->fill_iff[n],QiSio->words[n+3]);
        }
        break;
      } /* end case 'U' */
      
      case 'E': case 'e':
      {    /* end of data marker */
        has_eod = FOUND;
        
        /* check number of records */
   
        if(QiLineParser(QiSio) < 2){         /* parse line into tokens */
          QiErrorHandler(QMW_WARNING,"Badly formed EOD marker on .iff file");
          fprintf(logfile,"%s\n",&(QiSio->line[0]));
          has_eod = NOT_FOUND;
          check=QMW_WARNING;
          break;
        }
        n_expct = atoi( &(QiSio->words[1][0]) );
        if(n_expct != QiScdf->n_recs){
            sprintf(txt, "%ld records expected\n=> %ld records found",
                          n_expct,QiScdf->n_recs);
            QiErrorHandler(QMW_WARNING, txt);
            check=QMW_WARNING;
        }
        break;
      }
      
      default:
      {
           /* ignore other lines on preliminary scan */
      }
     } /* end switch */
   }  /* end while */
      
   if(has_eod == NOT_FOUND){
     QiErrorHandler(QMW_ERROR,
                    "No recognisable 'End of Data' marker in .iff file");
   }
   
   /* found n_recs records in .iff, rewind before reading file properly */  
   rewind(f);
   
   if(n_over > 0){
     sprintf(txt,
     "%ld records after 'End of Data' marker in .iff file ignored", n_over);
     QiErrorHandler(QMW_WARNING, txt);
     check=QMW_WARNING;
   }
   return check;
   
} /* end QiIFFScan  */


/***********************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  iff_to_cdf
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiIFFRead
*
* PURPOSE: Convert iff input lines to cdf output.
*
* DESCRIPTION: Reads lines from iff file and processes according
*              to content to populate the cdf file.
*              
*              
*
* ARGUMENTS:
*    IN:       FILE *f,                     pointer to iff file
*    IN:       struct QiSParams *QiScdf     structure with cdf quantities
*    IN:       struct QiSBuffer *QiSrec     structure for data buffer 
*                                           & data types
*    IN:       struct QiSIO *QiSio          structure to hold input line
*    IN:       struct QiSInterval *QiShi    structure for data interval
*                                           quantities
*
* GLOBALS:  
*
* RETURN:
*    TYPE:   long
*    VALUES: QMW_OK     Executed successfully.
*
* FUNCTIONS_USED:
*         QiErrorHandler
*         QiFillLine
*         QiLineParser
*         QiInterval
*
* ALGORITHM:
*            Read each iff file line in turn.
*            While line not NULL
*              Parse line (excepting messages, comments & blank lines)
*               If line is a data record: (R)
*                 skip lines with wrong number of words, or after EOD marker
*                 set buffer to space chars.
*                 extract record number, issue warning if different from 
*                 expected number
*                 skip text time string
*                 extract msec time tag and keep copy as last value found.
*                 Find time interval since last record if contiguous. 
*                 Add epoch of time tag to buffer = msec + start of day
*                 For each variable in list:
*                   If Epoch, skip as already done
*                   If VARY:
*                      for number of entries:
*                        convert next word on parsed list to appropriate 
*                        data type
*                        If IFF fillval change to cdf FILLVAL
*                        add to buffer
*                 Write buffer to cdf file
*                 increment record number
*               If line is Gattr (G)
*                 Use QiPutGattr to place Global attribute
*               If line is Vattr (V)
*                  Select attribute and variable from appropriate words
*                  Get attr data type
*                  write attribute value(s) from remaining word(s) 
*                  according to data type
*               If line is message, pass to logfile. (M)
*               If line is EOD (E)
*                 Set EOD flag.          
*               If line is Data Gap (N)
*                  If flag = GAP do nothing
*                  Else 
*                    set flag = GAP
*                    (= end of a data interval, start set by Record line) 
*                    write Gattr data_intervals pair 
*            
*
* HISTORY:
*
*   See also module prologue.
*
*        1 Dec 94    A.J.Allen
*        First Release
*     
***********************************************************************/
 
long QiIFFRead(FILE *f, 
             struct QiSParams *QiScdf, 
             struct QiSBuffer *QiSrec, 
             struct QiSIO *QiSio,
             struct QiSInterval *QiShi,
             double   data_res){
      
  long iff_flag=0;
  long line_mark;
  long recnum=0;
  long place=GAP;
  long post_gap=DATA;
  long varnum=0;
  long check;
  CDFstatus status;
  char txt[TEXT_LEN];
  double gap_tol;
  
  QiShi->ivl_min=0;
  QiShi->ivl_max=0;
  QiShi->ivl_ave=0;
  QiShi->ivl_num=0;
     
  QiScdf->n_bad = 0;

   /* set gap tolerance to 1.5 times expected interval */

   gap_tol = (3 * data_res) / 2;
  
  /* read each line of .iff in turn and pass extracted info to .cdf file */
     
  while(QiFillLine(QiSio, f) != NULL && iff_flag != END_OF_DATA){
      
    line_mark = QiSio->line[0];
    
    if(line_mark == 'M' || line_mark == '%' || line_mark == '\0' 
                        || line_mark == ' ' || line_mark == '\n'){
      /*  don't parse line as we don't process it */
    }
    else{
      /* parse line into tokens */
      if(QiLineParser(QiSio) == 0) {  
        QiErrorHandler(QMW_WARNING, "line unrecognised by QiLineParser");
        line_mark = '%';
      }
    }


    switch(line_mark){
        
      /* data record */
      case 'R': case 'r':
      {
        long test = QMW_OK;
        long word_ptr=0;
        static long record=-1;
        long n;
        double msec;
        size_t n_ptr = 0;
            
        /* test that `end of data' marker not passed */
        if(iff_flag == END_OF_DATA){
          /* skip this record as it is beyond the memory allocation */
          /* NOTE: warning message already generated by QiIFFScan module */
          break; 
        }
        
        /* test that line has correct number of data elements */
        if(QiSio->n_words - 3 != QiScdf->n_data){
          /* skip this record as it is incomplete */
          sprintf(txt, "Bad record length, after iff record %ld",record);
          QiErrorHandler(QMW_WARNING, txt);
          break; 
          
        }
            
        memset(QiSrec->buff, ' ', QiSrec->buff_len);
            
        word_ptr++;
        record = atoi(&(QiSio->words[word_ptr][0]));
                    
        /* skip time string  */
        word_ptr++;

        /* extract msec since start of day */
        ++word_ptr;
        msec = (double )atof(&(QiSio->words[word_ptr][0]));
         
       /* keep last time for SCALEMAX attribute; records are time ordered. */
        QiScdf->msec_last_rec = msec;

       /* calculate data intervals */
       
        check = QiInterval(QiShi, &place, &post_gap, recnum, msec, gap_tol);
          sprintf(txt,"At iff rec number %ld, time tags not monotonic?", 
                       record);
          if (check == QMW_ERROR) QiErrorHandler(check, txt); 
          
        /* set up EPOCH value and add to buffer */
        msec += QiScdf->msec_day; 
            
        memcpy((QiSrec->buff)+n_ptr, (char *)&msec, 8);
        n_ptr += 8;
             
        /* add all elements of each variable to buffer in turn */ 
            
        for(varnum=0; varnum < QiScdf->n_vars; varnum++){

          /* ignore EPOCH as already done */
          if(varnum == QiScdf->EPOCH_var_num) continue;

          if(QiScdf->recvary[varnum] == VARY){ /* ignore NOVARY variables */
                              
            for(n = 0; n < QiScdf->n_cells[varnum] ; n++){
                
              word_ptr++;
                  
              /* check for  fill value and convert to CDF value */
              if(QiScdf->n_fills == QiScdf->n_data){
                if(strcmp(QiSio->words[word_ptr], 
                          QiScdf->fill_iff[word_ptr - 3]) == 0) {
                   strcpy(QiSio->words[word_ptr], QiScdf->fill_cdf[varnum]);
                }
              }
                    
              switch(QiScdf->data_type[varnum]){

              case CDF_INT2: case CDF_UINT2:
              {
                QiSrec->sint_data = atoi(QiSio->words[word_ptr]);

                memcpy((QiSrec->buff)+n_ptr, (char *)&(QiSrec->sint_data),2);
                n_ptr += 2;
                break;
              }
              case CDF_INT4: case CDF_UINT4: 
              {
                QiSrec->int_data = atoi(QiSio->words[word_ptr]);
                    
                memcpy((QiSrec->buff)+n_ptr, (char *)&(QiSrec->int_data), 4);
                n_ptr += 4;
                break;
              }
              case CDF_REAL8: case CDF_DOUBLE: case CDF_EPOCH:
              {
                QiSrec->double_data = atof(QiSio->words[word_ptr]);
                                                 
                memcpy((QiSrec->buff)+n_ptr,(char *)&(QiSrec->double_data), 8);
                n_ptr += 8;
                break;
              }
              case CDF_REAL4: case CDF_FLOAT:
              {
                QiSrec->real_data = atof(QiSio->words[word_ptr]);
                    
                memcpy((QiSrec->buff)+n_ptr, (char *)&(QiSrec->real_data), 4);
                n_ptr += 4;
                break;
              }
              case CDF_CHAR: case CDF_UCHAR:
              {
                char UnHexStr[WORD_LEN+1];
                QiStrcpy(&(UnHexStr[0]), QiSio->words[word_ptr]);    
                if(QiScdf->n_chars[varnum] < (long) strlen(UnHexStr) ){
                  QiErrorHandler(QMW_ERROR, 
                                 "String variable longer than defined");
                }
                memcpy((QiSrec->buff)+n_ptr, &(UnHexStr[0]), strlen(UnHexStr));
                n_ptr += QiScdf->n_chars[varnum]; 
                    
              /* note buff initialised to white space at start of switch */
                break;
              }
              case CDF_UINT1: 
              {
                unsigned char one_byte;
                unsigned int int_byte;
                 
                int_byte = (unsigned int) QiStrToUInt(QiSio->words[word_ptr]);
                
                if(int_byte > 255 || int_byte < 0){  
                   sprintf(txt, 
                          "int*1 data type range? iff record=%ld, value=%d", 
                          record, int_byte );              
                   QiErrorHandler(QMW_ERROR, txt);
                }
                one_byte = int_byte;

                memcpy((QiSrec->buff)+n_ptr, (unsigned char *) &one_byte, 1);
                n_ptr++;
                break;
              }
              case CDF_BYTE: case CDF_INT1: 
              {
                char one_byte;
                int int_byte;
                 
                int_byte = (int) QiStrToInt(QiSio->words[word_ptr]);
                
                if(int_byte > 127 || int_byte < -128){  
                   sprintf(txt, 
                          "int*1 data type range? iff record=%ld, value=%d", 
                          record, int_byte );              
                   QiErrorHandler(QMW_ERROR, txt);
                }
                one_byte = int_byte;

                memcpy((QiSrec->buff)+n_ptr, (char *) &one_byte, 1);
                n_ptr++;
                break;
              }
              default:
              {
                sprintf(txt,"reading record: data type not implemented: %ld",
                        QiScdf->data_type[varnum]);
                QiErrorHandler(QMW_ERROR, txt);
              }
            } /* end switch */

          } /* end for over elements */
                          
        } /* end if on VARY  */
      } /* end for on variables */
                 
       if (test == QMW_WARNING) {
         sprintf(txt, "Badly formed record, last good iff rec was %ld",record);
         QiErrorHandler(QMW_WARNING, txt);
         QiScdf->n_bad++;
         break;
       }

       status = CDFlib(SELECT_, zVARs_RECNUMBER_, recnum, NULL_);
       status = CDFlib(PUT_, zVARs_RECDATA_, QiScdf->n_list, QiScdf->list_vars, 
                                              (void *)(QiSrec->buff), 
                        NULL_ );
         QiStatusHandler(status,"Problem writing record\n");
        
        recnum++;
        break;
      } /* end case 'R' */
          
      /* global attribute */
      case 'G': case 'g':
      {
        long check_g;
   
        check_g = QiPutGattr(QiSio, QiSrec);
        QiErrorHandler(check_g, "Problem returned from QiPutGattr function");
        break;

      } /* end case 'G' */
          
      /* End of data marker */
      case 'E': case 'e':
      {
        /* change flag so that we stop `while loop' next time */
        iff_flag = END_OF_DATA;
            
        /* find average half interval                         */
        
        if(QiShi->ivl_num != 0){ 
          QiShi->ivl_ave = 0.5 * QiShi->ivl_ave / QiShi->ivl_num;
        }
         
      } /* end case 'E': Fall through to 'N' for last entry in Data_interval */
          
      /* data gap, No data */
      case 'N': case 'n':
      {
                        
        if(place == GAP){
          /* do nothing, this is a continuation of the gap */
        }
        else{
           /* we are ENTERING a gap (so at least 1 record has been written) */
          place = GAP;
          
        }
        break;
      } /* end case 'N' */
                    
      /* variable attribute */
      case 'V': case 'v':
      {
        long n_in;
        long attr_type;
        long repeat;
        
        status = CDFlib(SELECT_, ATTR_NAME_,   QiSio->words[2],
                                 zENTRY_NAME_, QiSio->words[1],
                        NULL_);
        if (status != CDF_OK) {
          sprintf(txt,"can't locate variable=%s, attribute=%s",
                  QiSio->words[1], QiSio->words[2]);
          QiErrorHandler(QMW_WARNING, txt);
          break;
        }
        
        /* allow for multiple element attributes */
        
        repeat = QiSio->n_words - 3; 
          
          CDFlib(GET_, zENTRY_DATATYPE_, &attr_type, NULL_);
       
          switch(attr_type){
              case CDF_INT4: case CDF_UINT4: 
              {
                for(n_in=0 ; n_in < repeat; n_in++){
                  QiSrec->int_array[n_in] = atoi(QiSio->words[n_in+3]);
                }
                status=CDFlib(PUT_, zENTRY_DATA_,  
                                    attr_type, repeat, &(QiSrec->int_array[0]), 
                              NULL_);
                break;
              }
              case CDF_REAL8: case CDF_DOUBLE: case CDF_EPOCH:
              {
                for(n_in=0 ; n_in < repeat; n_in++){
                  QiSrec->double_array[n_in] = (double) atof(QiSio->words[n_in+3]);
                }
                status=CDFlib(PUT_, zENTRY_DATA_, 
                                    attr_type, repeat, &(QiSrec->double_array[0]), 
                              NULL_);
                break;
              }
              case CDF_REAL4: case CDF_FLOAT:
              {
                for(n_in=0 ; n_in < repeat; n_in++){
                  QiSrec->real_array[n_in] = atof(QiSio->words[n_in+3]);
                }
                status=CDFlib(PUT_, zENTRY_DATA_, 
                                    attr_type, repeat, &(QiSrec->real_array[0]), 
                              NULL_);
                break;
              }
              case CDF_INT2: case CDF_UINT2:
              {
                for(n_in=0 ; n_in < repeat; n_in++){
                  QiSrec->sint_array[n_in] = atoi(QiSio->words[n_in+3]);
                }
                status=CDFlib(PUT_, zENTRY_DATA_, 
                                    attr_type, repeat, &(QiSrec->sint_array[0]), 
                              NULL_);
                break;
              }
              case CDF_CHAR: case CDF_UCHAR:
              {
                 char UnHexStr[WORD_LEN];
                 
                  QiStrcpy(&(UnHexStr[0]), &(QiSio->words[3][0]));
                  n_in = strlen(&(UnHexStr[0]));
                  status=CDFlib(PUT_, zENTRY_DATA_,  
                                      attr_type, n_in, &(UnHexStr[0]),
                                NULL_);
                
                break;
              }
              case CDF_UINT1:
              {
                for(n_in=0 ; n_in < repeat; n_in++){
                  QiSrec->uchar_data[n_in] = (unsigned int) 
                                              QiStrToUInt(QiSio->words[n_in+3]);
                }
                status=CDFlib(PUT_, zENTRY_DATA_,  
                                     attr_type, repeat, 
                                     &(QiSrec->uchar_data[0]),
                              NULL_);
                break;
              }
              case CDF_BYTE:  case CDF_INT1:
              {
                for(n_in=0 ; n_in < repeat; n_in++){
                  QiSrec->char_data[n_in] = (int) 
                                            QiStrToInt(QiSio->words[n_in+3]);
                }
                status=CDFlib(PUT_, zENTRY_DATA_,  
                                     attr_type, repeat, &(QiSrec->char_data[0]),
                              NULL_);
                break;
              }
              default:
              {
                sprintf(txt,"writing V attr: data type not implemented: %ld\n",
                             attr_type);
                QiErrorHandler(QMW_ERROR, txt);
              }
              
              sprintf(txt,"Problem writing Vattr %s for %s value %s\n", 
                     QiSio->words[2], QiSio->words[1], QiSio->words[3]);
              QiStatusHandler(status,txt);
         } /* end switch */
        
        break;
      } /* end case 'V' */
          
      /* processing message */
      case 'M': case 'm':
      {
        /* pass message on to logfile for logging */
        fprintf(logfile, "%s\n", &(QiSio->line[0]));
        break;
      } /* end case 'M' */
          
      /* lines not used second scan */
      case '%': case 'U': case 'u': case 'T': case 't':
      case 'I': case 'i': case 'S': case 's':
      case '\0': case ' ': case '\n':
      {
        /* this is NULL 
        *  % comment lines are for .iff file clarity only 
        *  U, I and T lines are only read on first scan 
        *  empty lines are ignored                             
        */
        
        break;
      } /* end case NULL */
          
      /* bad lines */
      default:
      {
        fprintf(logfile, "\n bad line input in .iff file \n\n");
        fprintf(logfile, "%s\n\n", &(QiSio->line[0]) );
      }        
        
    } /* end switch on .iff input line content */
        
  } /* end while on reading .iff file */

   fprintf(logfile,"%ld records processed from iff file.\n", recnum);

   if(QiScdf->n_bad > 0){
     sprintf(txt, "%ld BAD DATA RECORDS in .iff file. These are ignored", 
                  QiScdf->n_bad);
     QiErrorHandler(QMW_WARNING, txt);
   }

  return QMW_OK;
} /* end QiIFFRead  */

/***********************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  iff_to_cdf
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiCtoISO
*
* PURPOSE: Converts time string  to Iso standard format.
*          
*
* DESCRIPTION: Converts time string from standard C library to Iso standard 
*              text format.
*
*              
* ARGUMENTS:
*    IN:       char *time                         input time string
*    OUT:      char *iso_time                     returned ISO time string
*
* GLOBALS:  
*
* RETURN:
*    TYPE:   char *
*    VALUES: Iso_time_string     
*
* FUNCTIONS_USED:
*         
*
* ALGORITHM:
*            Cut off time/date elements from back
*            convert text month to digital month
*            Assemble Iso time format from elements
*            
*
* HISTORY:
*
*   See also module prologue.
*
*        1 Dec 94    A.J.Allen
*        First Release
*     
***********************************************************************/
 
char * QiCtoISO(char *time, char *iso_time){

char year[5];
char month[3];
char mon[4];
char day[3];
char hour[9];
long n_char;

  /* remove newline from time and put year in year */
  time[24] = '\0';
  strcpy(year, &time[20]);
  
  time[7] = '\0';
  strcpy(mon, &time[4]);
  if(strcmp(mon,"Jan")==0){
    strcpy(month,"01");
  }
  else if(strcmp(mon,"Feb")==0){
    strcpy(month,"02");
  }
  else if(strcmp(mon,"Mar")==0){
    strcpy(month,"03");
  }
  else if(strcmp(mon,"Apr")==0){
    strcpy(month,"04");
  }
  else if(strcmp(mon,"May")==0){
    strcpy(month,"05");
  }
  else if(strcmp(mon,"Jun")==0){
    strcpy(month,"06");
  }
  else if(strcmp(mon,"Jul")==0){
    strcpy(month,"07");
  }
  else if(strcmp(mon,"Aug")==0){
    strcpy(month,"08");
  }
  else if(strcmp(mon,"Sep")==0){
    strcpy(month,"09");
  }
  else if(strcmp(mon,"Oct")==0){
    strcpy(month,"10");
  }
  else if(strcmp(mon,"Nov")==0){
    strcpy(month,"11");
  }
  else if(strcmp(mon,"Dec")==0){
    strcpy(month,"12");
  }
  else{
    mon[3] = '\0';
    strcpy(month, mon); /* as diagnostic */
  }
  time[10]='\0';
  strcpy(day, &time[8]);
  
  time[19]='\0';
  strcpy(hour, &time[11]);

 
  /* assemble output string */
  
  strcpy(iso_time,year);strcat(iso_time,"-");
  strcat(iso_time,month);strcat(iso_time,"-");
  strcat(iso_time,day);strcat(iso_time,"T");
  strcat(iso_time,hour);strcat(iso_time,".000Z");
    
  /* replace white space with 0 for ISO format */
  for (n_char=0; n_char<25; n_char++){
    if(iso_time[n_char]==' ') iso_time[n_char]='0';
  }
 
  return iso_time;
} /* end QiCtoISO  */


/***********************************************************************

* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  iff_to_cdf
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiFillLine
*
* PURPOSE: Read line from iff file.
*          
*
* DESCRIPTION: Reads lines of iff file one at a time.
*              Allocates sufficient memory for each read line.
*              Frees space from earlier calls before reading new line.
*
*              
* ARGUMENTS:
*    IN:       FILE *stream             File pointer to input iff file
*    OUT:      char **line              pointer to pointer of returned string
*
* GLOBALS:  
*
* RETURN:
*    TYPE:   char *
*    VALUES: pointer to line if successful
*            NULL if unsuccessful or EOF read.     
*
* FUNCTIONS_USED:
*         
*
* ALGORITHM:
*           Note position of file pointer.
*           read line one char at a time counting chars until \n or EOF
*           Return to file position at start of line
*           Use free and malloc to create large enough string
*           Use fgets to read line 
*
* HISTORY:
*
*   See also module prologue.
*
*        1 Dec 94    A.J.Allen
*        First Release
*     
***********************************************************************/

char * QiFillLine(struct QiSIO *io, FILE *stream){

 fpos_t posn;
 int ret;
 int n=0;
 int next;
 
 /* remember where we are */
  ret = fgetpos(stream, &posn);
  if(ret != 0) return NULL;
  
  while( !feof(stream) )
  {
    n++;
    next = fgetc(stream);
    if (next == '\n') break;
  }
  
  /* n is the number of chars excluding the '\0; */
  
  if (n == 0) return NULL;
  
  n++; /* allow for null terminator */
  
  /* go back and read line */
  fsetpos (stream, &posn);
  io->line = QiStrFree(io->line);
  io->line_cpy = QiStrFree(io->line_cpy);
  io->line     = (char *) malloc( n*sizeof(char) );
  io->line_cpy = (char *) malloc( n*sizeof(char) );
  
  if (fgets(io->line, n, stream) == NULL){
    return NULL;
  }
  return io->line;
   
} /* end QiFillLine */


/***********************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  iff_to_cdf
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiPutGattr
*
* PURPOSE: Write Global attribute to cdf file.
*          
*
* DESCRIPTION: Writes global attribute entry at next available entry
*              number. Will overwrite any entry that exactly matches the
*              string PLACEHOLDER.
*
*              
* ARGUMENTS:
*    IN:      struct QiSIO *QiSio          Input line parsed into words 
*    IN:      struct QiSBuffer *QiSrec     space for each data type
*
* GLOBALS:  
*
* RETURN:
*    TYPE:    long
*    VALUES:  QMW_OK     Executed successfully.    
*
* FUNCTIONS_USED:
*         CDFlib
*         QiStatusHandler
*
* ALGORITHM:
*           Determine number of elements to write from number of words in line
*           Select attribute.
*           determine data type of entry in prototype cdf
*           (csds Gattrs do not change data type between entries)
*           determine next unused entry number (from number values written)
*           If only one previous entry:
*               If previous entry = PLACEHOLDER
*                   Overwrite PLACEHOLDER
*           WRITE new Gattr entry (as array of n_elements)
*
* HISTORY:
*
*   See also module prologue.
*
*        1 Dec 94    A.J.Allen
*        First Release
*     
***********************************************************************/

long QiPutGattr(struct QiSIO *QiSio,
             struct QiSBuffer *QiSrec)
{
      
 long n_in;
 long attr_type;
 long n, n_elem, n_chars;
 CDFstatus status;
 char txt[TEXT_LEN];

            
   n_elem = QiSio->n_words - 2;
                
   status = CDFlib(SELECT_, ATTR_NAME_,  QiSio->words[1], NULL_);
   if (status != CDF_OK) { 
      sprintf(txt, "can't locate G attribute %s\n", QiSio->words[1]);
      QiStatusHandler(status, txt);
      return QMW_WARNING;
   }            
        
  /* initialise data type to that of first entry in .skt */

   status = CDFlib(SELECT_, gENTRY_,          0L,
                   GET_,    gENTRY_DATATYPE_, &attr_type, 
                   NULL_);
   sprintf(txt, "Problem getting data type of Gattr %s\n", QiSio->words[1]);
   QiStatusHandler(status, txt);
          
        /* we write sequentially, so the number of entries is */
        /* entry number at which to write next */ 
 
   status = CDFlib( GET_,  ATTR_NUMgENTRIES_, &n_in,   NULL_);
   sprintf(txt, "Problem getting number of Gattrs %s\n",  QiSio->words[1]);
   QiStatusHandler(status, txt);

       /*  if this is first entry, see if it is a text placeholder */
   if(n_in == 1){
     switch(attr_type){
               
      case CDF_CHAR: case CDF_UCHAR:
      {
        status=CDFlib( GET_,  gENTRY_DATA_, QiSrec->char_data,   NULL_);
        if (status != CDF_OK) {
          sprintf(txt,"Problem reading G attribute = %s", QiSio->words[1]);
          QiStatusHandler(status, txt);
          /* carry on if not fatal */
          break;
        } 
        /* append terminator as CDF does not! */
        status=CDFlib( GET_,  gENTRY_NUMELEMS_, &n_chars,   NULL_);
        QiSrec->char_data[n_chars] = '\0';

        if(strcmp(QiSrec->char_data, PLACEHOLDER) == 0){
          /* if placeholder, overwrite it with new entry */
          n_in = 0;
        }
        break;
      }
      default:
      {
            /* ignore other data types as we can't test for placeholder */
      }
     } /* end switch */
   }
            
        /* add new attribute entry */
            
   CDFlib( SELECT_, gENTRY_,   n_in,   NULL_);
            
   switch(attr_type){
     case CDF_INT2: case CDF_UINT2:
     {
        for (n = 0; n < n_elem; n++){
        QiSrec->sint_array[n] = atoi(QiSio->words[n+2]);
     }
        status=CDFlib(PUT_, gENTRY_DATA_,  attr_type, n_elem,
                      &(QiSrec->sint_array[0]),
                      NULL_);
        break;
     }
     case CDF_INT4: case CDF_UINT4: 
     {
        for (n = 0; n < n_elem; n++){
        QiSrec->int_array[n] = atoi(QiSio->words[n+2]);
     }
        status=CDFlib(PUT_, gENTRY_DATA_,  attr_type, n_elem, 
                      &(QiSrec->int_array[0]),
                      NULL_);
        break;
     }
     case CDF_REAL8: case CDF_DOUBLE: case CDF_EPOCH:
     {
        for (n = 0; n < n_elem; n++){
          QiSrec->double_array[n] = (double) atof(QiSio->words[n+2]);
        }
        status=CDFlib(PUT_, gENTRY_DATA_, attr_type, n_elem, 
                      &(QiSrec->double_array[0]),
                      NULL_);
        break;
     }
     case CDF_REAL4: case CDF_FLOAT:
     {
        for (n = 0; n < n_elem; n++){
          QiSrec->real_array[n] =  atof(QiSio->words[n+2]);
        }
        status=CDFlib(PUT_, gENTRY_DATA_, attr_type, n_elem, 
                      &(QiSrec->real_array[0]), 
                      NULL_);
         break;
     }
     case CDF_BYTE:  case CDF_INT1: 
     {         
        for (n = 0; n < n_elem; n++){
          QiSrec->char_data[n] =  QiStrToInt(&(QiSio->words[n+2][0]));
        }
         
         status=CDFlib(PUT_, gENTRY_DATA_,  attr_type, n_elem, 
                       &( QiSrec->char_data[0]),
                       NULL_);
         break;
     }
     case CDF_UINT1: 
     {         
        for (n = 0; n < n_elem; n++){
          QiSrec->uchar_data[n] =  QiStrToUInt(&(QiSio->words[n+2][0]));
        }

         status=CDFlib(PUT_, gENTRY_DATA_,  attr_type, n_elem, 
                       &(QiSrec->uchar_data[0]),
                       NULL_);
         break;
     }
     case CDF_CHAR: case CDF_UCHAR:
     {   
         char UnHexStr[WORD_LEN+1];
         
         QiStrcpy(&(UnHexStr[0]),&(QiSio->words[2][0]));      
         n_elem = strlen(UnHexStr);
         status=CDFlib(PUT_, gENTRY_DATA_,  attr_type, n_elem, 
                       &(UnHexStr[0]),
                       NULL_);
         break;
     }
     default:
     {
        sprintf(txt,"writing G attr: data type not implemented: %ld\n", 
                attr_type);
        QiErrorHandler(QMW_ERROR, txt);
     }
   } /* end switch */
            
   if (status != CDF_OK) {
     sprintf(txt,"Problem writing attribute %s, value %s\n", 
             QiSio->words[1], QiSio->words[2]);
     QiStatusHandler(status, txt);
   }            

   return QMW_OK;

 } /* end QiPutGattr */

/***********************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  iff_to_cdf
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiGetVarInfo
*
* PURPOSE: Get FILLVAL, data typa & data byte length for specified variable.
*          
*
* DESCRIPTION: Examines the prototype CDF file to determine the
*              csds value of FILLVAL for the variable. Fills appropriate
*              elements of an array corresponding to the variable number.   
*              Data type and associated byte size of the related variable
*              are determined as a by-product and stored.
*
*              
* ARGUMENTS:
*    IN/OUT:      struct QiSParams *QiScdf     cdf quantities 
*    IN:          struct QiSBuffer *QiSrec     space for each data type
*    IN:          char *varname                name of variable
*    IN:          long varnum                  id number of variable
*
* GLOBALS:  
*
* RETURN:
*    TYPE:    CDFstatus
*    VALUES:  Status code from CDFlib      
*
* FUNCTIONS_USED:
*         CDFlib
*         QiStatusHandler
*         QiErrorHandler
*
* ALGORITHM:
*           Select FILLVAL Vattr (variable already selected by calling
*           function)
*           Enquire data type
*           According to data type:
*               Set data length
*               Get FILLVAL
*           
*           
* HISTORY:
*
*   See also module prologue.
*
*        1 Dec 94    A.J.Allen
*        First Release
*     
***********************************************************************/

CDFstatus QiGetVarInfo (struct QiSParams *QiScdf, 
                        struct QiSBuffer *QiSrec,
                        char *varname,
                        long varnum){

 char txt[TEXT_LEN];
 CDFstatus status;

         /* find data type and FILLVAL for all rec-varying variables */
                             
         CDFlib(SELECT_, ATTR_NAME_,  "FILLVAL", 
                         zENTRY_,      varnum, 
                NULL_);

   /* NOTE: The fill value in the CDF file is only that of the .skt file if 
    *       the -fillval flag is set in the arguments to CDFskeleton. 
    *       FILLVAL will be of same data type as variable 
    */
       
         status = CDFlib(GET_, zVAR_DATATYPE_, &(QiScdf->data_type[varnum]),
                         NULL_);
           sprintf(txt, " Problem getting data type of variable %s", varname);
           QiStatusHandler(status, txt);
           
         switch(QiScdf->data_type[varnum]){
                case CDF_INT2: case CDF_UINT2:
                {
                  QiScdf->data_len[varnum] = 2;

                  /* keep fill value to be used in .cdf          */
                  /* & initialise value expected in .iff to same */

                  CDFlib(GET_, zENTRY_DATA_, (short *)&(QiSrec->sint_data),
                           NULL_);

                  sprintf(&(QiScdf->fill_cdf[varnum][0]), 
                              "%d", QiSrec->sint_data);

                  break;
                }
                case CDF_INT4: case CDF_UINT4: 
                {
                  QiScdf->data_len[varnum] = 4;
                    
                  /* keep fill value to be used in .cdf          */
                  /* & initialise value expected in .iff to same */
                    
                  CDFlib(GET_,    zENTRY_DATA_, (long *)&(QiSrec->int_data),
                           NULL_);

                  sprintf(&(QiScdf->fill_cdf[varnum][0]), 
                              "%ld", QiSrec->int_data);
                    
                  break;
                }
                case CDF_REAL8: case CDF_DOUBLE: case CDF_EPOCH:
                {
                  QiScdf->data_len[varnum] = 8;
                    
                  /* keep fill value to be used in .cdf          */
                  /* & initialise value expected in .iff to same */
                    
                  CDFlib(GET_,    zENTRY_DATA_, 
                         (double *)&(QiSrec->double_data),
                         NULL_);

                  sprintf(&(QiScdf->fill_cdf[varnum][0]), 
                            "%10.1g", QiSrec->double_data);
                  break;
                }
                case CDF_REAL4: case CDF_FLOAT:
                {
                  QiScdf->data_len[varnum] = 4;
                    
                   /* keep fill value to be used in .cdf  as text */
                   /* & initialise value expected in .iff to same */
                    
                  CDFlib(GET_,    zENTRY_DATA_, 
                               (float *)&(QiSrec->real_data),
                         NULL_);

                  sprintf(&(QiScdf->fill_cdf[varnum][0]), 
                               "%10.1e", QiSrec->real_data);
                  sprintf(&(QiScdf->fill_cdf[varnum][0]), "%10.1e", -10.e30);
                  break;
                }
                case CDF_BYTE:  
                case CDF_INT1:  
                case CDF_CHAR: 
                {
                  QiScdf->data_len[varnum] = 1;
                    
                  /* keep fill value to be used in .cdf          */
                  /* & initialise value expected in .iff to same */
                  
                  CDFlib(GET_, zENTRY_DATA_, (char *)&(QiSrec->char_data[0]),
                           NULL_);

                  strcpy(&(QiScdf->fill_cdf[varnum][0]), QiSrec->char_data); 

                  break;
                }

                case CDF_UINT1: 
                case CDF_UCHAR:
                {
                  QiScdf->data_len[varnum] = 1;
                    
                  /* keep fill value to be used in .cdf          */
                  /* & initialise value expected in .iff to same */
                    
                  CDFlib(GET_, zENTRY_DATA_,
                                 (unsigned char *)&(QiSrec->uchar_data[0]),
                           NULL_);
                  sprintf(&(QiScdf->fill_cdf[varnum][0]), 
                                   "%d", *(QiSrec->uchar_data) );
                  break;
                }
                default:
                {
                  sprintf(txt,"Get var info: data type not implemented: %ld\n", 
                               QiScdf->data_type[varnum]);
                  QiErrorHandler(QMW_ERROR, txt);
                }
         } /* end switch */
           
         return status;
           
} /* end QiGetVarInfo */

/***********************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  iff_to_cdf
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiWriteEpochAttrs
*
* PURPOSE: Write Vattrs associated with Epoch variable.
*          
*
* DESCRIPTION: Calculates and then writes 
*              VALIDMIN, VALIDMAX, SCALEMIN, SCALEMAX for Epoch.
*
*              
* ARGUMENTS:
*    IN:           struct QiSFName *QiSname    Structure for filename & 
*                                              substrings/values.
*    IN:          struct QiSParams *QiScdf     cdf quantities
*
* GLOBALS:  
*
* RETURN:
*    TYPE:    long
*    VALUES:  QMW_OK     Executed successfully.
*             QMW_ERROR  Fatal error detected.
*
* ALGORITHM:
*           Use QiComposeName to construct full Epoch name for this file.
*           (MIN and MAX values are set the same for VALID and SCALE)
*           MIN value determined from epoch of start of day. 
*               (Found in calling function).
*           Write VALIDMIN & SCALEMIN
*           MAX value determined as enclosing day boundary by:
*              convert epoch of last record to component elements
*              add 1 to day 
*              truncate to start of day.
*           Write VALIDMAX & SCALEMAX
*           
* HISTORY:
*
*   See also module prologue.
*
*        1 Dec 94    A.J.Allen
*        First Release
*     
***********************************************************************/

long QiWriteEpochAttrs(struct QiSFName *QiSname, struct QiSParams *QiScdf){

  CDFstatus status;
  long check=QMW_OK;
  char varname[CDF_VAR_NAME_LEN+1];
  double epoch_value;
  long year, month, day, hour, minute, second, msec;
  
   /* construct variable name */
 
  check = QiComposeName(varname, "Epoch", QiSname);
  QiErrorHandler(check, "failed to construct Epoch variable name");
  
 /* Epoch SCALE defaults as start and end of days enclosing data */
 
 /* construct minimum. First record must be after start of day in file name */
 
  epoch_value = QiScdf->msec_day;
 
 /* write min into SCALEMIN */
 
  status = CDFlib(SELECT_, ATTR_NAME_,  "SCALEMIN", 
                           zENTRY_NAME_, varname,
                           NULL_);
     QiStatusHandler(status,"can't locate V attribute `SCALEMIN' " );
     
  status = CDFlib(PUT_, zENTRY_DATA_, CDF_EPOCH, (long)1, &epoch_value, NULL_);
     QiStatusHandler(status,"can't write V attribute `SCALEMIN' " );
  
  /* find MAX. File may contain more than one day. */
  
  /* epoch of last record */
  epoch_value = QiScdf->msec_last_rec + QiScdf->msec_day;  
  EPOCHbreakdown(epoch_value, 
                 &year, &month, &day, &hour, &minute, &second, &msec);
  
   /*truncate to day boundary */
  epoch_value = QiGetDay0(year, month, day);     
 
 /* increment to next enclosing day */    
 
 epoch_value = epoch_value + 86400*1000;                 
    
  status = CDFlib(SELECT_, ATTR_NAME_,  "SCALEMAX",  NULL_);
     QiStatusHandler(status,"can't locate V attribute `SCALEMAX'\n" );

  status = CDFlib(PUT_, zENTRY_DATA_, CDF_EPOCH, (long)1, &epoch_value, NULL_);
     QiStatusHandler(status,"can't write V attribute `SCALEMAX' " );
      
     return check;
     
 } /* end QiWriteEpochAttrs */
 

/***********************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  iff_to_cdf
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiInterval
*
* PURPOSE: Determine data intervals between records.
*          
*
* DESCRIPTION: Keeps a running tally on intervals between time tags on
*              records. Allows generation of the Half_interval variable and
*              traps unmarked data gaps.
*
*              
* ARGUMENTS:
*    IN/OUT:   struct QiSInterval *QiShi   structure for data interval 
*                                          quantities
*    IN/OUT:   long *place                 flag for  DATA or GAP
*    IN/OUT:   long *post_gap              flag for first record post gap
*    IN/OUT:   long *interval              first and last rec number of 
*                                          interval
*    IN:       long recnum                 record number
*    IN:       double msec                 time tag of record (msec of day)
*
* GLOBALS:  
*
* RETURN:
*    TYPE:    long
*    VALUES:  QMW_OK     Executed successfully.
*             QMW_ERROR  Fatal error detected.
*
* FUNCTIONS_USED:
*         QiErrorHandler
*         QiStatusHandler
*
* ALGORITHM:
*         If first record after gap
*           set first of interval pair to this record
*           set start time to this time tag
*           set interval to FILLVAL (in case it is isolated data point)
*           set post_gap flag for identification of second record after gap
*         Else
*           find interval
*           add to running average
*           If 2nd record after gap 
*              reset first value to this (replace FILLVAL)
*              reset post_gap flag
*           Else
*              test for consistency of time tag spacing with previous interval
*              keep largest difference
*              if interval > 1.5*ave_interval, flag possible unmarked gap
*         Reset place flag to DATA  
*           
* HISTORY:
*
*   See also module prologue.
*
*        1 Dec 94    A.J.Allen
*        First Release
*     
***********************************************************************/

long QiInterval(struct QiSInterval *QiShi, 
                long *place, 
                long *post_gap,
                long recnum,
                double msec,
                double gap_tol){
 
  long check=QMW_OK;
  CDFstatus status;
  char txt[TEXT_LEN];
              
            /***********************************************************
            * NOTE: The following algorithm for finding the Half_interval      
            * assumes that the spin/measuring period is effectively  
            * constant over each segment of continuous data.    
            ***********************************************************/
            
        /* if this is the first data record after a gap... */
        if(*place == GAP){
            

          /* start measuring interval size */
          QiShi->ivl_start = msec;    
              
          /* initialise Half_interval for first record after gap to FILLVAL */
          QiShi->ivl_array[recnum] = IVL_FILL;
              
          /* set flag to identify second record after gap */     
          *post_gap = GAP;     
        }
        else{
              
          /* if not first data record after gap, find interval size */
          QiShi->ivl_array[recnum] = msec - QiShi->ivl_start;
          QiShi->ivl_start = msec;
          if(QiShi->ivl_min > QiShi->ivl_array[recnum]){
            QiShi->ivl_min = QiShi->ivl_array[recnum];
          }
          if(QiShi->ivl_max < QiShi->ivl_array[recnum]){
            QiShi->ivl_max = QiShi->ivl_array[recnum];
          }

         /* test for monotonicity in timing of records */
         if(QiShi->ivl_array[recnum] <= 0. ) {
           QiErrorHandler(QMW_ERROR,
                          "Bad time interval, records out of order");
         }
              
          /* trap unmarked data gaps */
          if(QiShi->ivl_array[recnum] > gap_tol){

             sprintf(txt,
              "Unmarked data gap at cdf rec %ld - Treating as new interval",
               recnum);
            /*   QiErrorHandler(QMW_WARNING, txt); */
             QiShi->ivl_array[recnum] = IVL_FILL;
             *post_gap = GAP;

             return QMW_WARNING;
           }

          /* everything is OK so add interval to running average */

          QiShi->ivl_ave += QiShi->ivl_array[recnum];
          QiShi->ivl_num++;

   /* if second record after gap, reassign value for first record to same */
          if(*post_gap == GAP){
            QiShi->ivl_array[recnum-1] = QiShi->ivl_array[recnum];
            *post_gap = DATA;
          }
           
        } /* end else on GAP */
            
        /* set flag that we are in a data interval */
        *place = DATA;
        
        return check;

} /* end QiInterval */

/***********************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  iff_to_cdf
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiStrcpy
*
* PURPOSE: Copy string and convert hex escapes to uchar.
*          
*
* DESCRIPTION: Copies string to new string character at a time, 
*              converting hex escapes to uchar values.
*              Sufficient space must be allocated elsewhere.
*              
* ARGUMENTS:
*    IN:   char * instring 
*                     
*    OUT:  char * outstring
*
* GLOBALS:  
*
* RETURN:
*    TYPE:    long
*    VALUES:  none
*
* FUNCTIONS_USED:
*
* ALGORITHM:
*         scan input string char by char. 
*         if '\\' detected test next char
*            if 'x' or 'X' convert the string \x?? to integer
*            write int value to next place in outstring
*         else add char to outstring
*         increment place in instring and outstring
*           
* HISTORY:
*
*   See also module prologue.
*
*        22 May 95    A.J.Allen
*        Added at V1.9
*     
***********************************************************************/

long QiStrcpy(char *outstring, char *instring) 
{
  long n;
  long n_out = 0;
  unsigned int ival;
  
  for (n=0; instring[n] != '\0'; n++){
    if(instring[n] == '\\' && (instring[n+1] == 'x' || instring[n+1] == 'X') ) {
        sscanf(&(instring[n+2]), "%2x", &ival );
        outstring[n_out] = ival;
        outstring[n_out+1] = '\0';
        n_out++;
        n += 3;
    }
    else{
      outstring[n_out] = instring[n];
      outstring[n_out+1] = '\0';
      n_out++;    
    }
  }

   return QMW_OK;             
   
} /* end QiStrcpy */

/**********************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  iff_to_cdf
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiCleanUp
*
* PURPOSE: Ensure an entry is prtesent in Inst_mode.
*
* DESCRIPTION: Empty files may have no Inst_mode set by PI software,
*              this is tested and corrected here.
*
* ARGUMENTS:
*   
*
* GLOBALS:  
*
* RETURN:
*    TYPE:   long
*    VALUES: QMW_OK      Executed successfully.
*            QMW_ERROR   Fatal error detected.
*
* FUNCTIONS_USED:
*         CDFlib
*         QiErrorHandler
*         QiStatusHandler
*         QiCtoISO
*
* ALGORITHM: 
*    
*        
* HISTORY:
*
*   See also module prologue.
*
*        28 Nov 95    A.J.Allen
*        First Release
*
***********************************************************************/

long QiCleanUp(struct QiSFName *QiSname, 
                       struct QiSParams *QiScdf)
{
  
  CDFstatus status;
  char mode_str[WORD_LEN];
  double epoch0 = QiScdf->msec_day;
  long index0 = 0;
  int stat0;
  unsigned char one_byte;
    
  /* Generated_by */
  
  status = CDFlib(SELECT_, ATTR_NAME_,  "Inst_mode", NULL_);
     QiStatusHandler(status,"can't locate G attribute `Inst_mode' " );

  CDFlib(SELECT_, gENTRY_, 0L,   
         GET_,  gENTRY_DATA_, &(mode_str[0]), NULL_);

 /* if file still contains PLACEHOLDER no G attr was written to Inst_mode */

  if (strncmp(mode_str, PLACEHOLDER, strlen(PLACEHOLDER)) == 0) {
    if(QiScdf->n_recs == 1){
      sprintf(mode_str,
              "%s>%s>0>%4.4ld-%2.2ld-%2.2ldT00:00:00Z>%s%s>%s",
               QiSname->sc,
               QiSname->iii,
               QiSname->year,
               QiSname->month,
               QiSname->day,
               QiSname->iii,
               NO_DATA,
               NO_DATA_TXT);
    }
    else{
      sprintf(mode_str,
              "%s>%s>0>%4.4ld-%2.2ld-%2.2ldT00:00:00Z>%s%s>%s",
               QiSname->sc,
               QiSname->iii,
               QiSname->year,
               QiSname->month,
               QiSname->day,
               QiSname->iii,
               NO_MODE,
               NO_MODE_TXT);
    }
    status = CDFlib(PUT_, gENTRY_DATA_, 
                          CDF_CHAR, (long)strlen(mode_str), mode_str, 
                    NULL_);
    QiStatusHandler(status,"failed to add entry to `Inst_mode' " );  
  }    
  
  /* ensure first record has valid Epoch value (FILLVAL is -ve) */
  
  status = CDFlib(SELECT_, zVAR_, QiScdf->EPOCH_var_num,
           SELECT_, zVAR_RECNUMBER_, 0L,   
           GET_,  zVAR_DATA_, &epoch0, NULL_);
           QiStatusHandler(status,"can't read epoch value at rec 0 " );

  if (epoch0 < 0.0){
    epoch0 = QiScdf->msec_day;
    status = CDFlib(PUT_,  zVAR_DATA_, &epoch0, NULL_);
         QiStatusHandler(status,"can't replace bad epoch value at rec 0 " );
  }
  
  return QMW_OK;
 
} /* end QiCleanUp */

/***********************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  iff_to_cdf
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiStrToUInt
*
* PURPOSE: convert str to unsigned int allowing for hex escapes.
*          
*
* DESCRIPTION: Converts str to unsigned int, 
*              converting hex escapes.
*              
* ARGUMENTS:
*    IN:   char * instring 
*                     
*
* GLOBALS:  
*
* RETURN:
*    TYPE:    long
*    VALUES:  none
*
* FUNCTIONS_USED:
*
* ALGORITHM:
*         scan input string char by char. 
*         if '\\' detected test next char
*            if 'x' or 'X' convert the string \x?? to integer
*            write int value to next place in outstring
*         else add char to outstring
*         increment place in instring and outstring
*           
* HISTORY:
*
*   See also module prologue.
*
*        14 Aug 97    A.J.Allen
*     
***********************************************************************/

unsigned int QiStrToUInt(char *instring) 
{
  unsigned int ival;

  
  
  if(instring[0] == '\\' && (instring[1] == 'x' || instring[1] == 'X') ) {
      sscanf(&(instring[2]), "%2x", &ival );
  }
  else{
      ival = (unsigned int) atoi(&(instring[0]));    
  }

   return ival;             
   
} /* end QiStrToUInt */

/***********************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT: QIFF
* 
* MODULE:  iff_to_cdf
*
* LANGUAGE: ANSI C
*
* FUNCTION: QiStrToInt
*
* PURPOSE: convert str to int allowing for hex escapes.
*          
*
* DESCRIPTION: Converts str to int, 
*              converting hex escapes.
*              
* ARGUMENTS:
*    IN:   char * instring 
*                     
*
* GLOBALS:  
*
* RETURN:
*    TYPE:    long
*    VALUES:  none
*
* FUNCTIONS_USED:
*
* ALGORITHM:
*         scan input string char by char. 
*         if '\\' detected test next char
*            if 'x' or 'X' convert the string \x?? to integer
*            write int value to next place in outstring
*         else add char to outstring
*         increment place in instring and outstring
*           
* HISTORY:
*
*   See also module prologue.
*
*        14 Aug 97    A.J.Allen
*     
***********************************************************************/

int QiStrToInt(char *instring) 
{
   int ival;

  
  
  if(instring[0] == '\\' && (instring[1] == 'x' || instring[1] == 'X') ) {
      sscanf(&(instring[2]), "%2x", &ival );
  }
  else{
      ival = (int) atoi(&(instring[0]));    
  }

   return ival;             
   
} /* end QiStrToInt */

/***********************************************************************/

char * QiStrFree( char * str){

  if ( str != NULL){
    free (str);
    str = NULL;
  }
  return str;

} /*  end QiStrFree */
